﻿// See https://aka.ms/new-console-template for more information
using GrapeCity.Documents.Drawing;
using GrapeCity.Documents.Imaging;
using GrapeCity.Documents.Pdf;
using GrapeCity.Documents.Pdf.Annotations;
using GrapeCity.Documents.Pdf.MarkedContent;
using GrapeCity.Documents.Pdf.Recognition;
using GrapeCity.Documents.Pdf.Recognition.Structure;
using GrapeCity.Documents.Pdf.Structure;
using GrapeCity.Documents.Pdf.TextMap;
using GrapeCity.Documents.Text;
using System;
using System.Collections.Generic;
using System.Drawing;
using System.IO;
using System.Text;

ExtractAllText();
ExtractPageText();
ExtractTextFromBounds();
ExtractImages();
ExtractRegex();
ExtractPDFInformation();
ExtractAttachments();
ExtractFormData();
ExtractTableData();
ExtractMultiPageTableData();
ExtractStructureTagData();
ExtractFonts();

void ExtractAllText()
{
    using (var fs = File.OpenRead(Path.Combine("Resources", "PDFs", "ItalyDoc.pdf")))
    {
        //Load Sample PDF document
        GcPdfDocument doc = new GcPdfDocument();
        doc.Load(fs);

        //Extract All text from PDF document
        var text = doc.GetText();
        //Display the results:
        Console.WriteLine("PDF Text: \n \n" + text);        
    }
}

void ExtractPageText()
{
    using (var fs = File.OpenRead(Path.Combine("Resources", "PDFs", "ItalyDoc.pdf")))
    {
        //Load Sample PDF document
        GcPdfDocument doc = new GcPdfDocument();
        doc.Load(fs);        

        //Extract specific page text:
        var pageText = doc.Pages[0].GetText();
        //Display the results:
        Console.WriteLine("PDF Page Text: \n" + pageText);
    }
}

void ExtractTextFromBounds()
{
    using (var fs = File.OpenRead(Path.Combine("Resources", "PDFs", "ItalyDoc.pdf")))
    {
        //Load Sample PDF document
        GcPdfDocument doc = new GcPdfDocument();
        doc.Load(fs);
        //Get Page TextMap
        var tmap = doc.Pages[0].GetTextMap();

        //Retrieve text at a specific (known to us) geometric location on the page:
        float tx0 = 7.1f, ty0 = 2.0f, tx1 = 3.1f, ty1 = 3.5f;
        HitTestInfo htiFrom = tmap.HitTest(tx0 * 72, ty0 * 72);
        HitTestInfo htiTo = tmap.HitTest(tx1 * 72, ty1 * 72);
        tmap.GetFragment(htiFrom.Pos, htiTo.Pos, out TextMapFragment range1, out string text);

        //Display the results:
        Console.WriteLine("Text extracted from specific bounds: \n \n" + text);
    }
}

void ExtractImages()
{
    using (var fs = File.OpenRead(Path.Combine("Resources", "PDFs", "ItalyDoc.pdf")))
    {
        //Load Sample PDF document
        GcPdfDocument doc = new GcPdfDocument();
        doc.Load(fs);

        //Extract images from the loaded PDF
        var imageInfos = doc.GetImages();

        int n = 1;
        foreach (var imageInfo in imageInfos)
        {
            //Save extracted images
            GcBitmap extImage = imageInfo.Image.GetNativeImage(Color.Transparent, Color.Transparent);
            extImage.SaveAsPng("Image" + n.ToString() + ".png");
            extImage.Dispose();
            n++;
        }
    }
}

void ExtractRegex()
{
    using (var fs = File.OpenRead(Path.Combine("Resources", "PDFs", "InvoiceDemo.pdf")))
    {
        //Load Sample PDF document
        GcPdfDocument doc = new GcPdfDocument();
        doc.Load(fs);

        //Find Invoice total amount
        FindTextParams searchParam1 = new FindTextParams(@"(Total)\r\n\$([-+]?[0-9]*\.?[0-9]+)", false, false, 72, 72, true, true);
        IList<FoundPosition> pos1 = doc.FindText(searchParam1);
        string totalAmount = pos1[0].NearText.Substring(pos1[0].PositionInNearText + pos1[0].TextMapFragment[0].Length).TrimStart();
        Console.WriteLine("Total amount found using regex in FindText method: " + totalAmount);

        //Find customer's email address from Invoice
        FindTextParams searchParam2 = new FindTextParams(@"[a-zA-Z0-9+_.-]+@[a-zA-Z0-9.-]+", false, false, 72, 72, true, true);
        IList<FoundPosition> pos2 = doc.FindText(searchParam2);
        string foundEmail = pos2[0].NearText.Substring(pos2[0].PositionInNearText, pos2[0].TextMapFragment[0].Length);
        Console.WriteLine("Email Address found using regex in FindText method: " + foundEmail);
    }
}

void ExtractPDFInformation()
{
    using (var fs = File.OpenRead(Path.Combine("Resources", "PDFs", "C1Olap-QuickStart.pdf")))
    {
        //Load Sample PDF document
        GcPdfDocument doc = new GcPdfDocument();
        doc.Load(fs);     

        //Extract PDF file information using the DocumentInfo property
        Console.WriteLine("Displaying PDF file information:\n \n");
        Console.WriteLine("Author: " + doc.DocumentInfo.Author);
        Console.WriteLine("Creator: " + doc.DocumentInfo.Creator);
        Console.WriteLine("Producer: " + doc.DocumentInfo.Producer);
        Console.WriteLine("Subject: " + doc.DocumentInfo.Subject);
        Console.WriteLine("Title: " + doc.DocumentInfo.Title);
        Console.WriteLine("Creation Date: " + doc.DocumentInfo.CreationDate);
        Console.WriteLine("Keywords: " + doc.DocumentInfo.Keywords);
    }
}

void ExtractAttachments()
{
    using (var fs = File.OpenRead(Path.Combine("Resources", "PDFs", "DocFileAttachment.pdf")))
    {
        //Load Sample PDF document
        GcPdfDocument doc = new GcPdfDocument();
        doc.Load(fs);

        //Get File Attachment using FileAttachmentAnnotation 
        //Get first page from document
        Page page = doc.Pages[0];
        //Get the annotation collection from pages
        AnnotationCollection annotations = page.Annotations;
        //Iterates the annotations
        foreach (AnnotationBase annot in annotations)
        {
            //Check for the attachment annotation
            if (annot is FileAttachmentAnnotation)
            {
                FileAttachmentAnnotation fileAnnot = (FileAttachmentAnnotation)annot;
                FileSpecification.FileInfo file = fileAnnot.File.File;
                //Extracts the attachment and saves it to the disk
                string path = Path.Combine("Attachments", file.FileName);
                FileStream stream = new FileStream(path, FileMode.Create);
                file.EmbeddedFileStream.GetStream().CopyTo(stream);
                stream.Dispose();
            }
        }

        //Get Document Attachments (document level attachments )
        //Iterates through the attachments
        if (doc.EmbeddedFiles.Count != 0)
        {
            foreach (KeyValuePair<string, FileSpecification> attachment in doc.EmbeddedFiles)
            {
                //Extracts the attachment and saves it to the disk
                FileSpecification.FileInfo file = attachment.Value.File;
                string path = Path.Combine("Attachments", file.FileName);
                FileStream stream = new FileStream(path, FileMode.Create);
                file.EmbeddedFileStream.GetStream().CopyTo(stream);
                stream.Dispose();
            }
        };
    }
}

void ExtractFormData()
{
    using (var fs = File.OpenRead(Path.Combine("Resources", "PDFs", "filledForm.pdf")))
    {
        //Load Sample PDF document
        GcPdfDocument doc = new GcPdfDocument();
        doc.Load(fs);

        MemoryStream stream = new MemoryStream();
        //Export the form data to a stream   
        doc.ExportFormDataToXML(stream);

        //Alternatively, we can even export to a file of appropriate format
        //Export the form data to an XML file
        doc.ExportFormDataToXML("sampleFormData.xml");
    }
}

void ExtractFonts()
{
    using (var fs = File.OpenRead(Path.Combine("Resources", "PDFs", "ItalyDoc.pdf")))
    {
        //Load Sample PDF document
        GcPdfDocument doc = new GcPdfDocument();
        doc.Load(fs);

        //Extract list of fonts used in loaded PDF document
        var fonts = doc.GetFonts();

        Console.WriteLine($"Total of {fonts.Count} fonts found in PDF file: \n");

        int i = 1;
        foreach (var font in fonts)
        {            
            Console.WriteLine($"{i}. BaseFont: {font.BaseFont}; IsEmbedded: {font.IsEmbedded}.");
            ++i;
        }
    }
}

void ExtractTableData()
{
    using (var fs = File.OpenRead(Path.Combine("Resources", "PDFs", "zugferd-invoice.pdf")))
    {
        const float DPI = 72;
        const float margin = 36;
        var doc = new GcPdfDocument();

        var tf = new TextFormat()
        {
            FontSize = 9,
            ForeColor = Color.Black
        };
        var tfHdr = new TextFormat(tf)
        {
            FontSize = 11,
            ForeColor = Color.DarkBlue
        };
        var tfRed = new TextFormat(tf) { ForeColor = Color.Red };

        //The approx table bounds:
        var tableBounds = new RectangleF(0, 3 * DPI, 8.5f * DPI, 3.75f * DPI);

        var page = doc.NewPage();
        page.Landscape = true;
        var g = page.Graphics;

        var tl = g.CreateTextLayout();
        tl.MaxWidth = page.Bounds.Width;
        tl.MaxHeight = page.Bounds.Height;
        tl.MarginAll = margin;
        tl.DefaultTabStops = 150;
        tl.LineSpacingScaleFactor = 1.2f;

        var docSrc = new GcPdfDocument();
        docSrc.Load(fs);

        var itable = docSrc.Pages[0].GetTable(tableBounds);

        if (itable == null)
        {
            tl.AppendLine($"No table was found at the specified coordinates.", tfRed);
        }
        else
        {
            tl.Append($"\nThe table has {itable.Cols.Count} column(s) and {itable.Rows.Count} row(s), table data is:", tfHdr);
            tl.AppendParagraphBreak();
            for (int row = 0; row < itable.Rows.Count; ++row)
            {
                var tfmt = row == 0 ? tfHdr : tf;
                for (int col = 0; col < itable.Cols.Count; ++col)
                {
                    var cell = itable.GetCell(row, col);
                    if (col > 0)
                        tl.Append("\t", tfmt);
                    if (cell == null)
                        tl.Append("<no cell>", tfRed);
                    else
                        tl.Append(cell.Text, tfmt);
                }
                tl.AppendLine();
            }
        }

        TextSplitOptions to = new TextSplitOptions(tl) { RestMarginTop = margin, MinLinesInFirstParagraph = 2, MinLinesInLastParagraph = 2 };
        tl.PerformLayout(true);
        while (true)
        {
            var splitResult = tl.Split(to, out TextLayout rest);
            doc.Pages.Last.Graphics.DrawTextLayout(tl, PointF.Empty);
            if (splitResult != SplitResult.Split)
                break;
            tl = rest;
            doc.NewPage().Landscape = true;
        }
        doc.Save("ExtractTableData.pdf");
    }
}

void ExtractMultiPageTableData()
{
    const float DPI = 72;
    const float margin = 36;
    var doc = new GcPdfDocument();

    var tf = new TextFormat()
    {
        FontSize = 9,
        ForeColor = Color.Black
    };
    var tfHdr = new TextFormat(tf)
    {
        FontSize = 11,
        ForeColor = Color.DarkBlue
    };
    var tfRed = new TextFormat(tf) { ForeColor = Color.Red };

    using (var fs = File.OpenRead(Path.Combine("Resources", "PDFs", "product-list.pdf")))
    {
        var page = doc.NewPage();
        page.Landscape = true;
        var g = page.Graphics;

        var tl = g.CreateTextLayout();
        tl.MaxWidth = page.Bounds.Width;
        tl.MaxHeight = page.Bounds.Height;
        tl.MarginAll = margin;
        tl.DefaultTabStops = 165;

        var docSrc = new GcPdfDocument();
        docSrc.Load(fs);

        for (int i = 0; i < docSrc.Pages.Count; ++i)
        {
            // TableExtractOptions allow you to fine-tune table recognition accounting for
            // specifics of the table formatting:
            var teo = new TableExtractOptions();
            var GetMinimumDistanceBetweenRows = teo.GetMinimumDistanceBetweenRows;
            // In this particular case, we slightly increase the minimum distance between rows
            // to make sure cells with wrapped text are not mistaken for two cells:
            teo.GetMinimumDistanceBetweenRows = (list) =>
            {
                var res = GetMinimumDistanceBetweenRows(list);
                return res * 1.2f;
            };
            var top = i == 0 ? DPI * 2 : DPI;
            // Get the table at the specified bounds:
            var itable = docSrc.Pages[i].GetTable(new RectangleF(DPI * 0.25f, top, DPI * 8, DPI * 10.5f - top), teo);

            // Add table data to the text layout:
            tl.Append($"\nTable on page {i + 1} of the source document has {itable.Cols.Count} column(s) and {itable.Rows.Count} row(s), table data is:", tfHdr);
            tl.AppendParagraphBreak();
            for (int row = 0; row < itable.Rows.Count; ++row)
            {
                var tfmt = row == 0 ? tfHdr : tf;
                for (int col = 0; col < itable.Cols.Count; ++col)
                {
                    var cell = itable.GetCell(row, col);
                    if (col > 0)
                        tl.Append("\t", tfmt);
                    if (cell == null)
                        tl.Append("<no cell>", tfRed);
                    else
                        tl.Append(cell.Text, tfmt);
                }
                tl.AppendLine();
            }
        }

        // Print the extracted data:
        TextSplitOptions to = new TextSplitOptions(tl) { RestMarginTop = margin, MinLinesInFirstParagraph = 2, MinLinesInLastParagraph = 2 };
        tl.PerformLayout(true);
        while (true)
        {
            var splitResult = tl.Split(to, out TextLayout rest);
            doc.Pages.Last.Graphics.DrawTextLayout(tl, PointF.Empty);
            if (splitResult != SplitResult.Split)
                break;
            tl = rest;
            doc.NewPage().Landscape = true;
        }

        doc.Save("ExtractMultiPageTableData.pdf");
    }
}

void ExtractStructureTagData()
{
    using (var fs = File.OpenRead(Path.Combine("Resources", "PDFs", "TaggedPdf.pdf")))
    {
        var doc = new GcPdfDocument();
        doc.Load(fs);

        // Get the LogicalStructure and top parent element:
        LogicalStructure ls = doc.GetLogicalStructure();
        Element root = ls.Elements[0];

        foreach (Element e in root.Children)
        {
            string type = e.StructElement.Type;
            if (string.IsNullOrEmpty(type) || !type.StartsWith("H"))
                continue;
            // Note: topmost level is 1:
            if (!int.TryParse(type.Substring(1), out int headingLevel) || headingLevel < 1)
                continue;

            // Get the element text:
            string text = e.GetText();
            text = new string(' ', (headingLevel - 1) * 2) + text;
            Console.WriteLine(text);
        }
    }
}